# Copyright (c) 2008 Oracle Corporation.  All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions
#are met:
#
#     - Redistributions of source code must retain the above
#       copyright notice, this list of conditions and the following
#       disclaimer.
#     - Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials
#       provided with the distribution.
#     - Neither the name of Oracle Corporation nor the names of its
#       contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#POSSIBILITY OF SUCH DAMAGE.
#
#Alternatively, the contents of this file may be used under the terms
#of the GNU General Public License version 2 (the "GPL") distributed
#with this softwarere in the file COPYING.GPL, in which case the
#provisions of the GPL are applicable instead of the above.
#
#If you wish to allow the use of your version of this file only under
#the terms of the GPL and not to allow others to use your version of
#this file under the license above, indicate your decision by deleting
#the provisions above and replace them with the notice and other
#provisions required by the GPL.  If you do not delete the provisions
#above, a recipient may use your version of this file under the above
#license or the GPL.

import pexpect, os, sys, socket, time, mimetypes, re
from SocketServer import BaseServer
from BaseHTTPServer import HTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler
from OpenSSL import SSL
from threading import Thread, RLock

class SecureHTTPServer(HTTPServer, Thread):
	def __init__(self, server_address, HandlerClass, keyfile, certificate):
		BaseServer.__init__(self, server_address, HandlerClass)
		Thread.__init__(self)
		ctx = SSL.Context(SSL.SSLv23_METHOD)
		ctx.use_privatekey_file (keyfile)
		ctx.use_certificate_file(certificate)
		self.socket = SSL.Connection(ctx, socket.socket(self.address_family, self.socket_type))
		self.server_bind()
		self.server_activate()

	def run(self):
		print "Serving forever"
		self.serve_forever()
		print "Should not get here"

class SecureHTTPRequestHandler(SimpleHTTPRequestHandler):
	def setup(self):
		self.connection = self.request
		self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
		self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)
		self.workingdir = '/tmp/p2v'

	def do_GET(self):
		if self.path.endswith("reboot"):
			datafile = open("%s/%s" %(self.workingdir, self.path), 'r')
			self.send_response (200)
			self.send_header ("Content-type", "text/plain")
			self.end_headers ()
			while 1:
				c = datafile.read(1024)
				if c:
					self.wfile.write(c)
				else:
					datafile.close()
					break;
			os.popen4('/sbin/reboot')
			sys.exit(0)
		elif self.path.endswith(".img"):
	                # I would have used os.path.getsize to determine the size of the file
                	# but for whatever reason that returns 0 for block devices.  So instead
                	# I open and seek to the end and then call f.tell()
                	f = open("%s/%s" %(self.workingdir, self.path), 'r')
                	f.seek(0,2)
                	size = f.tell()
                	f.close()

			imagefile = os.path.basename(self.path)
			regex = re.compile("xvd[a-z]*|hd[a-d]|sd[a-z]*")
			m = regex.search(imagefile)

			if m:
				datafile = open("/dev/%s" %(m.group()), 'r')
			else:
				regex = re.compile("c[0-9]*d[0-9]*")
				m = regex.search(imagefile)
				datafile = open("/dev/cciss/%s" %(m.group()), 'r')

			self.send_response (200)
			self.send_header ("Content-type", "application/octet-stream")
			self.send_header ("Content-length", size)
			self.end_headers ()
			while 1:
				c = datafile.read(32768)
				if c:
					self.wfile.write(c)
				else:
					datafile.close()
					break;
		else:
			SimpleHTTPRequestHandler.do_GET(self)

	extensions_map = mimetypes.types_map.copy()
	extensions_map.update({
		'' : 'text/plain',
		'.img' : 'application/octet-stream',
		'.cfg' : 'text/plain',
	})

'''Utility function that will generate the ssl keys using the openssl command.
Defaults to creating the files in /tmp using the hostname returned by socket.gethostname().
Defaults can be overriden by providing the appropriate value:

p2vkey          the private key filename
p2vrequest      the request filename
p2vcert         the certificate filename
hostname        the hostname

Note that the contents of the files will be overwritten.

Returns the tuple (p2vkey, p2vrequest, p2vcert).
'''
def generateSSLkeys():
	# Some defaults
	p2vkey = '/tmp/p2vkey'
	p2vrequest = '/tmp/p2vrequest'
	p2vcert = '/tmp/p2vcertificate'
	hostname = socket.gethostname()

	# Find the openssl binary based on whether we are in a normal or install time environment
	opensslbin = ''
	if os.access('/usr/bin/openssl', os.X_OK):
		opensslbin = '/usr/bin/openssl'
	else:
		opensslbin = '/mnt/runtime/usr/bin/openssl'

	# Find the openssl.cnf file based on whether we are in a normal or install time environment
	if os.access('/etc/pki/tls/openssl.cnf', os.R_OK):
		cnffile = '/etc/pki/tls/openssl.cnf'
	else:
		cnffile = '/mnt/runtime/etc/pki/tls/openssl.cnf'

        # Generate the private key file
        (dummy, std_out_and_err) = os.popen4('%s genrsa -out %s 1024' %(opensslbin, p2vkey))

        # This is lame, but if I do not sleep for a bit between between the private key creation
        # and the request creation the request creation will intermittenly fail (1 out of 3 or so)
        time.sleep(1)

        # Generate the request.  All questions are left blank save the hostname.
        child = pexpect.spawn('%s req -new -key %s -out %s -config %s' %(opensslbin, p2vkey, p2vrequest, cnffile))
        child.setecho(False)
        child.expect('Country Name')
        child.sendline('.')
        child.expect('State or Province Name')
        child.sendline('.')
        child.expect('Locality Name')
        child.sendline('.')
        child.expect('Organization Name')
        child.sendline('.')
        child.expect('Organizational Unit Name')
        child.sendline('.')
        child.expect('Common Name')
        child.sendline(hostname)
        child.expect('Email Address')
        child.sendline('.')
        child.expect('A challenge password')
        child.sendline('.')
        child.expect('An optional company name')
        child.sendline('.')
        child.read()

        # Generate the certificate
        (dummy, std_out_and_err) = os.popen4('%s x509 -req -days 1000 -in %s -signkey %s -out %s' %(opensslbin, p2vrequest, p2vkey, p2vcert))

        # Return the key, request, and certificate files as a tuple
        return (p2vkey, p2vrequest, p2vcert)
